#include<cmath>
#include<cstdlib>
#include<cstdio>
#include "realtype.h"
#include "normaldist_mex.h"
#include "tauchen3.h"
#include "mex.h"

// #define DEBUG_IDENTITY_MATRIX

// #define DEBUG_tauchen2

int tauchen1d(REAL_TYPE *x, REAL_TYPE *Px, const REAL_TYPE mu, const REAL_TYPE rho, const REAL_TYPE sigma, const int N, const REAL_TYPE xlb, const REAL_TYPE xub)
// Model:    x(t+1) = mu + rho x(t) + sigma*e(t+1)
// evenly spaced grid between xlb and xub
// the space for x and Px should be already allocated
{

    if (abs(rho)<1 && sigma>0 && N>=2 && xlb<mu/(1-rho) && xub>mu/(1-rho))
    {

        REAL_TYPE dx = (xub - xlb)/((REAL_TYPE)N - 1.0);
        REAL_TYPE x1, x2, mean_x;
        x[0] = xlb;
        for (int i = 1; i<N; i++)
        {
            x[i] = xlb + ((REAL_TYPE)i)*dx;
        }

        for (int i = 0; i < N; i++)
        {
            mean_x = mu + rho*x[i];
            for (int j = 0; j < N; j++)
			{
                if (j==0)
                {
                    x2 = x[0] + 0.5*dx;
                    Px[i*N + j] = normcdf(x2 - mean_x, sigma);
                }
                else if (j==N-1)
                {
                    x1 = x[N-1] - 0.5*dx;
                    Px[i*N + j] = 1 - normcdf(x1 - mean_x, sigma);
                }
                else
                {
                    x1 = x[j] - 0.5*dx;
                    x2 = x[j] + 0.5*dx;
                    Px[i*N + j] = normcdf(x2 - mean_x, sigma) - normcdf(x1 - mean_x, sigma);
                }
            }
        }

        // check Px is a valid probability matrix
        bool bValid = true;
        REAL_TYPE sum;

        for (int i = 0; i < N; i++)
        {
            sum = 0;
            for (int j = 0; j < N; j++)
            {
                if (Px[i*N+j]<0)
                {
                    bValid = false;
                }
                sum += Px[i*N+j];
            }
            if (abs(1.0 - sum)>1e-10)
            {
                bValid = false;
            }
        }

        if (bValid)
        {
            return EXIT_SUCCESS;
        }
        else
        {
            mexPrintf("tauchen1d: did not produce valid transition probabilities.\n");
            return EXIT_FAILURE;
        }

    }
    else
    {
		mexPrintf("tauchen1d: check input parameters.\n");
        return EXIT_FAILURE;
    }

}

int ContructGrids(REAL_TYPE *z, REAL_TYPE *x, REAL_TYPE *Pz, REAL_TYPE *Px, REAL_TYPE *Pzx, REAL_TYPE *Qzx, const REAL_TYPE rhoz, const REAL_TYPE sigmaz, 
	const REAL_TYPE meanx, const REAL_TYPE rhox, const REAL_TYPE sigmax, const REAL_TYPE corr_shock, const int Nz, const int Nx, const REAL_TYPE stdwidth_z, const REAL_TYPE stdwidth_x)
// see the notes
// assumes grids are all created
{

	// check input parameters
	if (abs(rhoz)<1 && abs(rhox)<1 && sigmaz>0 && sigmax>0 && abs(corr_shock)<1 && stdwidth_z>0 && stdwidth_x>0 && Nz>=2 && Nx>=2)
	{

		REAL_TYPE zlb, zub, xlb, xub, uncondstd_z, uncondstd_x;

		uncondstd_z = sigmaz / sqrt(1 - rhoz*rhoz);
		uncondstd_x = sigmax / sqrt(1 - rhox*rhox);

		zlb = -1 * uncondstd_z*stdwidth_z;
		zub = +1 * uncondstd_z*stdwidth_z;
		xlb = meanx - uncondstd_x*stdwidth_x;
		xub = meanx + uncondstd_x*stdwidth_x;

		tauchen1d(z, Pz, 0.0, rhoz, sigmaz, Nz, zlb, zub);

		tauchen1d(x, Px, meanx*(1-rhox), rhox, sigmax, Nx, xlb, xub);

		// construct risk neutral prob. states ordered as (z,x)
		int iz, ix, jz, jx;
		REAL_TYPE SCALING;

		for (int i = 0; i < Nz*Nx; i++)
		{

			ix = i % Nx;
			iz = (i - ix) / Nx;

			// SCALING constant
			SCALING = 0;
			for (jz = 0; jz < Nz; jz++)
			{
				SCALING += exp(-x[ix]*corr_shock*(z[jz] - rhoz*z[iz]) + 0.5*x[ix]*x[ix]*(1 - corr_shock*corr_shock))*Pz[iz*Nz + jz];
			}

			SCALING = 1.0 / SCALING;

			for (int j = 0; j < Nz*Nx; j++)
			{
				jx = j % Nx;
				jz = (j - jx) / Nx;
				// physical probs
				Pzx[i*Nz*Nx + j] = Pz[iz*Nz + jz] * Px[ix*Nx + jx];
				// RN probs
				Qzx[i*Nz*Nx + j] = SCALING*exp(- x[ix] * corr_shock*(z[jz] - rhoz*z[iz]) + 0.5*x[ix] * x[ix] * (1 - corr_shock*corr_shock)) * Pz[iz*Nz + jz] * Px[ix*Nx + jx];
			}

		}

		return EXIT_SUCCESS;

	}
	else
	{
		mexPrintf("ContructGrids:: check parameters\n");
		return EXIT_FAILURE;
	}

}

int ContructGrids_efficient(REAL_TYPE *z, REAL_TYPE *x, REAL_TYPE *Pz, REAL_TYPE *Px, REAL_TYPE *Qzx_znext, const REAL_TYPE rhoz, const REAL_TYPE sigmaz, 
	const REAL_TYPE meanx, const REAL_TYPE rhox, const REAL_TYPE sigmax, const REAL_TYPE corr_shock, const int Nz, const int Nx, const REAL_TYPE stdwidth_z, const REAL_TYPE stdwidth_x)
// Qzx_znext is indexed as (z,x,znext)
{
    
    // check input parameters
	if (abs(rhoz)<1 && abs(rhox)<1 && sigmaz>0 && sigmax>0 && abs(corr_shock)<=1 && stdwidth_z>0 && stdwidth_x>0 && Nz>=2 && Nx>=1)
	{

		REAL_TYPE zlb, zub, xlb, xub, uncondstd_z, uncondstd_x;

		uncondstd_z = sigmaz / sqrt(1 - rhoz*rhoz);
		uncondstd_x = sigmax / sqrt(1 - rhox*rhox);

		zlb = -1 * uncondstd_z*stdwidth_z;
		zub = +1 * uncondstd_z*stdwidth_z;
		xlb = meanx - uncondstd_x*stdwidth_x;
		xub = meanx + uncondstd_x*stdwidth_x;

		tauchen1d(z, Pz, 0.0, rhoz, sigmaz, Nz, zlb, zub);

		if (Nx>=2)
		{
			tauchen1d(x, Px, meanx*(1-rhox), rhox, sigmax, Nx, xlb, xub);
		}
		else
		{
			// Nx==1
			x[0] = meanx;
			Px[0] = 1.0;
		}

		// construct risk neutral prob. states ordered as (z,x)
		int iz, ix, jz;
		REAL_TYPE SCALING;

		for (int i = 0; i < Nz*Nx; i++)
		{

			ix = i % Nx;
			iz = (i - ix) / Nx;

			// SCALING constant
			SCALING = 0;
			for (jz = 0; jz < Nz; jz++)
			{
				SCALING += exp(-x[ix]*corr_shock*(z[jz] - rhoz*z[iz]) + 0.5*x[ix]*x[ix]*(1 - corr_shock*corr_shock))*Pz[iz*Nz + jz];
			}

			SCALING = 1.0 / SCALING;

			for (int j = 0; j < Nz; j++)
			{
				// RN probs
				Qzx_znext[i*Nz + j] = SCALING*exp(- x[ix] * corr_shock*(z[j] - rhoz*z[iz]) + 0.5*x[ix] * x[ix] * (1 - corr_shock*corr_shock)) * Pz[iz*Nz + j];
			}

		}

		return EXIT_SUCCESS;

	}
	else
	{
		mexPrintf("ContructGrids:: check parameters");
		return EXIT_FAILURE;
	}

}